// ignore_for_file: cancel_subscriptions, public_member_api_docs
// ignore_for_file: non_constant_identifier_names, empty_catches
// ignore_for_file: avoid_dynamic_calls

import 'dart:async';
import 'dart:developer';

import 'package:flutter_blue_plus/flutter_blue_plus.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:permission_handler/permission_handler.dart';

class BLEManager with Ecg {
  factory BLEManager() {
    return _instance;
  }

  factory BLEManager.WithName({String name = 'HEH001'}) {
    return _instanceList[name] ?? _instance;
  }

  factory BLEManager.CardioClick() {
    return BLEManager.WithName(name: 'CardioClickV1');
  }

  BLEManager._internal() {}

  static Map<String, BLEManager> _instanceList = {
    'HEH001': _instance,
    'CardioClickV1': _instanceCardioClickV1,
  };

  static final BLEManager _instance = BLEManager._internal();
  static final BLEManager _instanceCardioClickV1 = BLEManager._internal();

  BluetoothDevice? device;

  final FlutterBluePlus _flutterBlue = FlutterBluePlus.instance;

  /// [points] is a map to register/unregister characteristics.
  /// To register characteristics,
  /// calling a [registerCharacteristics] function
  /// before starting any measurement.
  final points = UUID.characteristic.map(
    (key, value) {
      return MapEntry<Characteristics, BluetoothCharacteristic?>(
        key,
        null,
      );
    },
  );

  /// [streams] provides observable streams.
  /// It is a wrapper of state management library of mobX.
  ///
  /// It notifies changes of each state whether measurement is started or stop
  final streams = Streams();

  bool Function(BluetoothDevice device) scanFilter =
      (device) => device.localName.isWearEcg12;

  /// [connectDevice] tries to connect a BLE device
  /// and returns its connectivity.
  ///
  /// When putting a identifier of preferred device as an input,
  /// it attempts to connect the preferred device.
  ///
  /// If there are multiple devices discovered,
  /// connection priority is decided depends on the rssi.
  Future<bool> connectDevice({String preferredDeviceId = ''}) async {
    // 0. 초기화
    device = null;

    // await _flutterBlue.turnOn();

    await [
      Permission.bluetoothScan,
      Permission.bluetooth,
      Permission.bluetoothConnect,
      Permission.bluetoothAdvertise,
    ].request();

    // 1. 연결 장비 확인
    final connectedDevices = await _flutterBlue.connectedDevices;
    final connectedFeasibleDevices = connectedDevices.where(scanFilter);
    if (connectedFeasibleDevices.isNotEmpty) {
      device = connectedFeasibleDevices.firstWhere(
        (device) => device.remoteId.toString() == preferredDeviceId,
        orElse: () => connectedFeasibleDevices.first,
      );
      return true;
    }

    // 2. 스캔 수행
    // await _flutterBlue.stopScan();
    final scan = _flutterBlue.startScan(
      timeout: const Duration(seconds: 2),
    );

    final isFoundFeasibleDevice = await scan.then(
      (events) async {
        final scanResults = events as List<ScanResult>;
        final feasibleScanResult = scanResults
            .where((result) => scanFilter(result.device))
            .toList(growable: false)
          ..sort(
            (a, b) => b.rssi - a.rssi,
          );

        if (feasibleScanResult.isNotEmpty) {
          final result = feasibleScanResult.firstWhere(
            (r) => r.device.remoteId.toString() == preferredDeviceId,
            orElse: () => feasibleScanResult.first,
          );
          device = result.device;
          // if (Platform.isAndroid) {
          //   await device!.pair();
          // }
          await device!.connect();
          return true;
        }
        return false;
      },
    );

    return isFoundFeasibleDevice;
  }

  /// [disconnectDevice] disconnects all devices
  /// which are matched with the given scan filter
  Future<void> disconnectDevice() async {
    final connected = await _flutterBlue.connectedDevices;

    Future.wait(
      [
        for (final device in connected.where(scanFilter)) device.disconnect(),
      ],
    );
    device = null;
  }

  /// [registerCharacteristics] registers all characteristics,
  /// defined in [Characteristics].
  ///
  /// The result is registered in a map structure [points].
  Future<void> registerCharacteristics() async {
    if (device == null) return;

    final services = await device!.discoverServices();
    for (final service in services) {
      final characteristics = service.characteristics;
      for (final characteristic in characteristics) {
        UUID.characteristic.forEach(
          (key, value) {
            final uuid = characteristic.characteristicUuid.toString();
            if (uuid.toLowerCase() == value.toLowerCase()) {
              log('${key.name} was registered');
              points[key] = characteristic;
            }
          },
        );
      }
    }

    _sessionTime = DateTime.now();

    streams
      ..connectionStateStream = _getConnectionStateStream().broadcast
      ..batteryLevelStream = _getBatteryLevelStream().broadcast
      ..tapButtonStream = _getPhysicalTapStream().broadcast;
  }

  /// [initializeStreams] is used to reset measurement.
  ///
  /// Once calling it,
  /// all measurement streams become unobservable.
  Future<void> initializeStreams() async {
    streams
      ..rawStream = null
      ..decompressedStream = null
      ..recentStream = null
      ..leadStateStream = null
      ..heartRateStream = null
      ..bufferedStreamEcg = null
      ..bufferedFilterStreamEcg = null
      ..bufferedPlotStreamEcg = null
      ..bufferedPlotRTStreamEcg = null;
  }

  Stream<bool?> _getConnectionStateStream() async* {
    final stream = device?.connectionState;
    if (stream == null) return;

    FlutterBluePlus.instance.connectedDevices;

    var previousEmitTime = DateTime.now();
    BluetoothConnectionState? prev;

    final subscription = stream.listen((event) {
      prev = event;
      previousEmitTime = DateTime(0);
    });

    while (device != null) {
      final timeAfterLastEmit = DateTime.now().difference(previousEmitTime);
      if (timeAfterLastEmit > const Duration(seconds: 2)) {
        yield prev == BluetoothConnectionState.connected;
        previousEmitTime = DateTime.now();
      }
      await Future<void>.delayed(const Duration(seconds: 2));
    }
    yield null;

    await subscription.cancel();
  }

  Stream<int?> _getBatteryLevelStream() async* {
    final battery = points[Characteristics.HEH_BATTERY_LEVEL];
    if (battery == null) return;

    try {
      await battery.setNotifyValue(true);
    } catch (e) {
      log('_getBatteryLevelStream: $e');
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await battery.setNotifyValue(true);
      } catch (e) {
        log('_getBatteryLevelStream (second try): $e');
      }
    }

    final read = await battery.read();
    var prev = read.isNotEmpty ? read[0] : null;
    yield prev;

    var previousEmitTime = DateTime.now();
    final subscription = battery.value.listen((event) {
      final changed = event.isNotEmpty ? event[0] : null;
      prev = changed;
      previousEmitTime = DateTime(0);
    });

    while (device != null) {
      final timeAfterLastEmit = DateTime.now().difference(previousEmitTime);
      if (timeAfterLastEmit > const Duration(seconds: 5)) {
        yield prev;
        previousEmitTime = DateTime.now();
      }
      await Future<void>.delayed(const Duration(seconds: 5));
    }

    yield null;

    await subscription.cancel();
  }

  Stream<bool?> _getPhysicalTapStream() async* {
    final tap = points[Characteristics.HEH_SENSOR_LOCATION];
    if (tap == null) return;
    try {
      await tap.setNotifyValue(true);
    } catch (e) {
      log('_getPhysicalTapStream: $e');
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await tap.setNotifyValue(true);
      } catch (e) {
        log('_getPhysicalTapStream (second try): $e');
      }
    }

    final sessionTimeInside = _sessionTime;

    await for (final event in tap.value) {
      if (sessionTimeInside != _sessionTime) break;
      if (event.isEmpty) continue;
      yield true;
    }
    yield null;
  }

  DateTime? _sessionTime;
}
